/* wagon-ESP-OSC
 *  un wagon radio-guidé
 *  plateforme ESP8266 ESP-12
 *  commande moteur Olimex BB-L298
 *  protocole réseau OSC
 *  SNHack gepeto@du-libre.org
 *  Codes sous licences GPLV3
 */

/* projet LOCO-DOM
 *  IP du réseau 192.168.0.x/24
 *  les loco sont de x= 81 a 91
 *  ports 9081 a 9091
 * Connexions ESP-12 BBL-298
 * |-                |   ESP-12   |          |        -|
   |-                |      RST   |       TXD|        -|
   |LDR              | ADC        |       RXD|       - |
   |-                |CH_P        |    GPIO4 |  1 RFID |
   |4 BBL            |   GPIO16   |    GPIO5 |        -|
   |3 BBL            |   GPIO14   |    GPIO0 |        -|
   |led bleu2 BBL-298|   GPIO12   |    GPIO2 |        -|
   |led verte        |   GPIO13   |   GPIO15 |led rouge|
   |-                |   VCC 3V   |       GND|        -|
 *
 */

// WIFI
#include <ESP8266WiFi.h>
#include <WiFiUdp.h>
#include <Ticker.h> // gestion des timer

//MDNSResponder mdns;
// OSC
#include <OSCBundle.h>
#include <OSCMessage.h>
#include <OSCBoards.h>
//port numbers for OSC
const unsigned int inPort = 9081;
//destination IP pour la trace OSC
IPAddress MonitIP(192, 168, 0, 99); // IP et port du moniteur
const unsigned int MonitPort = 9998; // suivi normal
IPAddress DebugIP(192, 168, 0, 123); // IP et port de debug
const unsigned int DebugPort = 9999;
WiFiUDP udpin;
WiFiUDP udpout;
Ticker zou; // OSC envoi position
Ticker init_rfid; // init port rfid
// serial
#include <EspSoftSerialRx.h>
EspSoftSerialRx rfid;
const char* ssid     = "LOCOMO";
const char* password = "0123456789";
char nomhost[20]  = "LOCO-1";
// DHCP donc pas besoin
//IP
//IPAddress (192,168,IPC,IPD);
int IPC = 0;
int IPD = 81;
String nom = "LOCO-1";
char* OSCVar = "/LOCO-1";

// connect motor controller pins to Arduino digital pins
// ESP moteur traction
int enA = 12;
int in1 = 13;
int in2 = 14;
int Vitesse_actuelle = 0 ; // vitesse actuelle, <0 = arriere
uint32_t temps_boucle_total = 0 ; // en milli seconde
uint32_t temps_a_faire = 0;
uint32_t temps_depart = 0;
uint32_t temps_inter = 0;
uint32_t temps_tot = 0;
int rampe = 50;

const int nbr_max_de_puce = 40;
int nbr_de_puce = 0;
int no_puce_actuelle = 98;
int puce_a_trouver = 98;
int puce_derniere_lue = 98;
// tableau des puces rfid 20x13
char *RFID[nbr_max_de_puce * 13] = {
  "6D0070007964", // puce 0
  "010611B52F8C", // 1
  "0102BBAEF1E7", // 2
  "010445C652D4", // 3
  "010445D771E6", // 4
  "010445D8138B", // 5
  "0106116C7309", // 6
  "010611B118BF", // 7
  "05004E236C04", // 8
  "05004E23F79F", // 9
  "05004D1DE6B3", // 10
  "05004D0FEBAC", // 11
  "05004E23630B", // 12
  "1123456ab1",
  "1123456ab2",
  "1123456ab3",
  "1123456ab4",
  "1123456ab5",
  "1123456ab6",
  "1123456ab7",
  "1123456ab8",
};


static int32_t sequencenumber = 0;
int commande = 0; // no de la commande en cours

void setup()
{
  Serial.begin(115200);
  // set all the motor control pins to outputs
  pinMode(enA, OUTPUT);
  pinMode(in1, OUTPUT);
  pinMode(in2, OUTPUT);
  // serial RFID
  rfid.begin(9600, 2); // sortie GPIO2
  Serial.println(ESP.getChipId(), HEX);
  //setup ethernet part
  Serial.println();
  Serial.print("Connecting to ");
  Serial.println(ssid);
  WiFi.mode(WIFI_STA);
  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }
  //  DHCP donc pas fixé ici
  //WiFi.config(IPAddress(192, 168, IPC, IPD), IPAddress(192, 168, IPC, IPD), IPAddress(255, 255, 255, 0));
  Serial.println("");
  Serial.println("WiFi connected");
  Serial.println(ssid);
  Serial.println("IP address: ");
  Serial.println(WiFi.localIP());

  Serial.println("host: ");
  Serial.println(WiFi.hostname());
  WiFi.hostname(nomhost);
  Serial.println(WiFi.hostname());
  Serial.println("Starting UDP");
  udpin.begin(inPort);
  Serial.print("Local port: ");
  Serial.println(udpin.localPort());

  delay(100);
  // envoi le bonjour
  MonitMsg("BONJOUR", (int32_t)ESP.getChipId());
  // placer jesuisla en boucle timer 1s
  zou.attach(5, jesuisla);
  // init_rfid.attach(10, rfid_port_init); ?
  // En Avant;
  digitalWrite(in1, HIGH);
  digitalWrite(in2, LOW);
  //Vitesse_moteur(1000);

}
void rfid_port_init() { // re initialisation du port rfid les 10s
  Serial.println("init rfid");
  rfid.service();
}
void jesuisla() { // envoie OSC /msg if position vitesse
  OSCMessage zouzou(OSCVar);
  zouzou.add("tic");
  zouzou.add((int32_t)commande); // commande en cours
  zouzou.add((int32_t)no_puce_actuelle);
  zouzou.add((int32_t)Vitesse_actuelle);
  zouzou.add((int32_t)(millis() / 1000));
  udpout.beginPacket(MonitIP, MonitPort);
  zouzou.send(udpout);
  udpout.endPacket();
  udpout.beginPacket(DebugIP, DebugPort);
  zouzou.send(udpout);
  udpout.endPacket();
  zouzou.empty(); // free space occupied by
}
void Arriere() {
  digitalWrite(in1, LOW);
  digitalWrite(in2, HIGH);
}
void Avant() {
  digitalWrite(in1, HIGH);
  digitalWrite(in2, LOW);
}
char *Lit_serie(int taille) {
  char tagValue[taille];
  unsigned char val;
  Serial.println("lit_serie");

  for (int i = 0; i <= 11; i++) {
    if (rfid.read(val)) {
      tagValue[i] = '\0'; // Null
      // on saute avec le start 0x02 et stop bit 0x03
      if ((val == 0x0D) || (val == 0x0A) || (val == 0x03) || (val == 0x02)) {
        break;
      }
      Serial.print(i);
      Serial.print(" ");
      Serial.print((char)val);
      Serial.print(" ");
      Serial.println(val, HEX );
      tagValue[i] = (char)val;
      if ((i == 11) ) { // 11 digit read is complete
        tagValue[12] = '\0'; // Null
        //        Serial.print("Tag read: ");
        MonitMsg(tagValue, 0);
        return tagValue;
      }
    }
  }
}
int Lit_rfid() {
  // mise en forme de la lecture rfid
  //retourne le no_puce

  //Serial.println("attente rfid");
  int i = 0;

  char tagValue[12];
  for (i = 0; i <= 11; i++) {
    unsigned char val;
    if (rfid.read(val)) {
      tagValue[i] = '\0'; // Null
      // on saute avec le start 0x02 et stop bit 0x03
      if ((val == 0x0D) || (val == 0x0A) || (val == 0x03) || (val == 0x02)) {
        break;
      }
      //      Serial.print(i);
      //      Serial.print(" ");
      //      Serial.print((char)val);
      //      Serial.print(" ");
      //    Serial.println(val, HEX );
      tagValue[i] = (char)val;

      if ((i == 11) ) { // 11 digit read is complete
        tagValue[12] = '\0'; // Null
        //Serial.println(tagValue);
        int tagId = CherchePuce( tagValue );
        // Only fire the strike plate if this tag was found in the database
        if ( tagId != 99 )
        {
          //Serial.print("rfid puce ");
          //Serial.println(tagId);
          MonitMsg("puce ", tagId);
          return (tagId);
        } else {
          MonitMsg("puce inconnue", 99);
        }
      }
    }
  }
}

int CherchePuce( char tag[12] ) {
  for (int puce = 0; puce < nbr_de_puce; puce++) {
    // Check if the tag value matches this row in the tag database
    Serial.println(RFID[puce]);
    if (strcmp(tag, RFID[puce]) == 0)
      return (puce);
  }
  // Si pas trouvé on retourne 99
  return (99);
}


bool MonitMsg(char debugbla[], int debugint) {
  OSCMessage envoi(OSCVar);
  envoi.add(debugbla);
  envoi.add(debugint);

  udpout.beginPacket(MonitIP, MonitPort);
  envoi.send(udpout);
  udpout.endPacket();

  udpout.beginPacket(DebugIP, DebugPort);
  envoi.send(udpout);
  udpout.endPacket();
  envoi.empty();
}
int Vitesse_moteur(int moteur) {
  int i = 0;
  if (moteur > 1024) // filtre le max
    moteur = 1024;

  if (moteur < -1024)
    moteur = -1024;

  MonitMsg("vitesse", moteur);
  Serial.println(moteur);
  Serial.println(Vitesse_actuelle);
  if (Vitesse_actuelle > moteur) {
    for (i = Vitesse_actuelle; i >= moteur; i = i - 20) {
      if ( i >= 0 ) {
        //Avant;
        digitalWrite(in1, HIGH);
        digitalWrite(in2, LOW);
      } else {
        //Arriere;
        digitalWrite(in1, LOW);
        digitalWrite(in2, HIGH);
      }
      analogWrite(enA, abs(i));
      delay(50); // rampe
    }
  } else {
    for ( i = Vitesse_actuelle; i <= moteur; i = i + 20) {
      if ( i >= 0 ) {
        //Avant;
        digitalWrite(in1, HIGH);
        digitalWrite(in2, LOW);
      } else {
        //Arriere;
        digitalWrite(in1, LOW);
        digitalWrite(in2, HIGH);
      }
      analogWrite(enA, abs(i));
      delay(rampe); // rampe
    }
  }
  Vitesse_actuelle = moteur;
  analogWrite(enA, abs(moteur));
  no_puce_actuelle = 97;
}
// COMMANDES
void ArretMoteur(OSCMessage &msg) {
  commande = 0;
  Serial.println("ArretMoteur /stop");
  MonitMsg("STOP", 0);
  analogWrite(enA, 0);
}
void MoteurControl(OSCMessage &msg) {
  // met le moteur à la vitesse reçue
  commande = 1;
  Serial.println("/moteur");
  int moteur_lu = msg.getInt(0);
  Vitesse_moteur(moteur_lu);
}
/*
bool InitTableauPuces(OSCMessage &msg) {
  int i = 0;
  int puce = 0;
  commande = 7;
  nbr_max_de_puce = msg.getInt(0);
  MonitMsg("INIT Tableau", nbr_max_de_puce);
  Serial.println("/inittableaupuce");
  // tour pour nbr de puce total
  unsigned char val;
  char tagValue[12];
  MonitMsg("Apprendre puce 1", 1 );
  RFID[1] =  Lit_serie(12);// premiere puce 1
  for (puce = 2; puce <= nbr_max_de_puce; puce++) { // On doit déjà connaitre la puce 1
    strncpy(tagValue, Lit_serie(12), 12) ;
    if (strcmp(RFID[1], RFID[puce]) == 0) {
      nbr_de_puce = puce;
      MonitMsg("nbr de puce", puce);
      break;
    }
    strncpy(RFID[puce], tagValue, 12) ;
  }
}
*/
void Initialisation(OSCMessage &msg) {
  // la loco fait un tour complet doucement pour connaitre le nombre de puce
  // la premiere etant la puce 1
  // ou on lui mpresente les puces une par une
  // puis la loco fait un tour complet pour connaitre le temps nécessaire
  // et vérifier la lecture des puces RFID , voir faire l'init du tableau?
  // elle ne connait au départ que la puce 0, position 0
  // on ecrit le tableau sur le fs
  int i = 0;
  int puce = 0;
  unsigned char val;
  char tagValue[12];
  i = msg.getInt(0);
  if (i > 1) {
    nbr_de_puce = msg.getInt(0);
  }
  Serial.print("/init");
  Serial.println(nbr_max_de_puce);
  MonitMsg("INIT", nbr_max_de_puce);
  commande = 2;
  Vitesse_moteur(0);
  MonitMsg("Apprendre puce 1", 1 );
  Vitesse_moteur(650);
  RFID[1] =  Lit_serie(12);// premiere puce 1
  for (puce = 2; puce <= nbr_de_puce; puce++) { // On doit déjà connaitre la puce 1
    strncpy(tagValue, Lit_serie(12), 12) ;
    if (strcmp(RFID[1], RFID[puce]) == 0) {
      nbr_de_puce = puce;
      Vitesse_moteur(0);
      MonitMsg("nbr de puce", puce);
      break;
    }
    strncpy(RFID[puce], tagValue, 12) ;
  }
  Aller(1, -650); //retourner sur la puce 1 et y stopper
  uint32_t temps = (uint32_t)millis();
  //rouler a fond
  Aller(1, 1000);
  for (puce = 2; puce <= nbr_de_puce ; puce++) {
    if (strcmp(RFID[0], RFID[puce]) == 0) {
      break;
    }
    strncpy(tagValue, Lit_serie(12), 12) ;
    MonitMsg(RFID[puce], puce);
  }
  temps_boucle_total = (uint32_t)millis() - temps ;
  MonitMsg("temps boucle", temps_boucle_total / 1000);
  // enregistrer les parametres
  commande = 0;
  Aller(1, -650); //retourner sur la puce 1 et y stopper
}
void Rendezvous(OSCMessage &msg) {       // A FINIR
  // /goto ii puce temps : se rendre a la puce xx en tant de temps
  commande = 3;
  Serial.println("/rdv");
  puce_a_trouver = (int) msg.getInt(0);
  if (puce_derniere_lue != puce_a_trouver) {
    temps_a_faire = (uint32_t)msg.getInt(1) * 1000; // en s
    temps_depart = (uint32_t)millis();
    Vitesse_moteur((int)msg.getInt(1));
  }
}
void SeRendre(OSCMessage &msg) {
  // /aller ii puce vitesse : se rendre a la puce xx a n vitesse
  commande = 4;
  Aller((int)msg.getInt(0), (int)msg.getInt(1));
}
void LitPuce(OSCMessage &msg) {
  commande = 6;
  zou.detach();
  MonitMsg("lit", 0);
}
bool Aller(int no_puce, int vitesse) {
  // aller jusqu'a la puce no_puce a vitesse
  Serial.println("Aller");
  MonitMsg("aller", no_puce);
  puce_a_trouver = no_puce;
  temps_depart = millis();
  temps_inter = 0;
  temps_tot = 0;
  // if (no_puce == no_puce_actuelle) { // on refait un tour si encore
  //   MonitMsg("trouve", no_puce);
  //   return true;
  // }
  Vitesse_moteur(vitesse);
  return (true);
}
void loop()
{
  // Lire la comande et executer le debut
  OSCMessage msgIN;
  int size;
  if ( (size = udpin.parsePacket()) > 0)
  {
    while (size--)
      msgIN.fill(udpin.read());

    if (!msgIN.hasError()) {
      msgIN.dispatch("/stop", ArretMoteur);     // stop le moteur immediatement commande=0
      msgIN.dispatch("/moteur", MoteurControl); // met le moteur à la vitesse lue  commande=1
      msgIN.dispatch("/init", Initialisation);  // fait un tour pour reperer les puces (puce0 fixe) commande=2
//      msgIN.dispatch("/puces", InitTableauPuces);  // on présente les puces pour remplir le tableau commande=7
      msgIN.dispatch("/goto", Rendezvous);      // va à la puce dans le temps donné commande=3
      msgIN.dispatch("/aller", SeRendre);       // aller a la puce a une vitesse /aller ii puce vitesse commande=4
      msgIN.dispatch("/lit", LitPuce);       // renvoi le code de la puce lue  /lit  commande=6
    }
    udpin.flush();
  }
  // Lire une puce et continuer une commande en cours
  if (commande != 2) {
    if (commande == 6) {
      Lit_serie(12);
      delay(100);
    } else {
      int puce_lue = Lit_rfid();
      if (puce_lue > 0) {
        MonitMsg("passe", puce_lue);
        temps_inter = (uint32_t)millis() - temps_depart ;
        temps_tot += temps_inter;
        MonitMsg("en ", temps_inter / 1000);
        if ((commande == 4 || commande == 3) && puce_lue == puce_a_trouver) {
          //stopmoteur
          analogWrite(enA, 0);
          MonitMsg("puce trouve" , puce_lue);
          MonitMsg("en ", temps_tot / 1000);
          temps_tot = 0;
          no_puce_actuelle = puce_a_trouver;
          puce_derniere_lue = puce_a_trouver;
          puce_a_trouver = 98;
          commande = 0;

          delay(100);
        }
        no_puce_actuelle = puce_lue;
      }
    }
  }
  delay(20);
}
